/*******************************************************************************
 *
 * FILENAME:         I2CMux_1_to_1.v
 * Component Name:   I2CMux_1_to_1
 *
 * DESCRIPTION:
 * 
 *******************************************************************************
 *                 I*O Signals:
 *******************************************************************************
 *    Name              Direction       Description
 *
 *******************************************************************************
 * Copyright 2008-2012, Cypress Semiconductor Corporation.  All rights reserved.
 * You may use this file only in accordance with the license, terms, conditions,
 * disclaimers, and limitations in the end user license agreement accompanying
 * the software package with which this file was provided.
 ******************************************************************************/
 `include "cypress.v"

module I2CMux_1_to_1 
(
	input   wire    CLOCK,          /* 16x SCL Clock = 1.6 MHz @ 100 kHz */
    input   wire    M_SCL_IN,       /* Upstream master  SCL input */
    input   wire    M_SDA_IN,       /* Upstream master  SDA input */
    input   wire    S_SCL_IN,       /* Downstream slave SCL input */
	input   wire    S_SDA_IN,       /* Downstream slave SDA input */
    output  reg     M_SCL_OUT,      /* Upstream master  SCL output */
    output  reg     M_SDA_OUT,      /* Upstream master  SDA output */
    output  reg     S_SCL_OUT,      /* Downstream slave SCL output */
    output  reg     S_SDA_OUT,      /* Downstream slave SDA output */
    output  wire    DEBUG_1,        /* Debug output */
    output  wire    DEBUG_2,        /* Debug output */
    output  wire    DEBUG_3,        /* Debug output */
    output  wire    DEBUG_4         /* Debug output */
);

    /**************************************************************************/
    /* Internal Signals                                                       */
    /**************************************************************************/
    reg		        m_scl_d;        /* Used for master SCL edge detection */
	reg		        m_sda_d;        /* Used for master SDA edge detection */
	reg		        rising_m_scl;   /* Used for master SCL edge detection */
    reg             falling_m_scl;  /* Used for master SCL edge detection */
    reg             s_scl_d;        /* Used for slave SCL edge detection */
	reg		        i2c_start;      /* Set upon I2C start or repeated start */
    wire            m_stop;         /* Instantaneous master bus decodes */
    wire            m_start;        /* Instantaneous master bus decodes */
    wire            s_nack;         /* Instantaneous slave bus decodes */
    wire    [6:0]   bit_cnt;        /* I2C transfer bit counter */
	wire	        bit_tc;         /* I2C transfer bit counter tc */
    reg     [2:0]   scl_cnt;		/* Counter to measure width of SCL */
    reg             ack_time;       /* ACK time flag */
    reg             back_to_back;   /* Back-to-back transaction flag */
    reg             addr_phase;     /* Address phase flag */
    reg             read_cycle;     /* Read cycle flag */
    reg             addr_nack;      /* Slave nacks address flag */
    reg             upstream_sync;  /* Sync to upstream bus flag */
    
    /**************************************************************************/
    /* R&D Debug Outputs                                                      */
    /**************************************************************************/
    assign DEBUG_1 = addr_nack;
    assign DEBUG_2 = back_to_back;
    assign DEBUG_3 = ack_time;
    assign DEBUG_4 = upstream_sync;

    /**************************************************************************/
    /* Bus State Detection Logic                                              */
    /**************************************************************************/
	always @(posedge CLOCK) 
    begin
		m_scl_d       <=  M_SCL_IN;
		m_sda_d       <=  M_SDA_IN;
        s_scl_d       <=  S_SCL_IN;
		rising_m_scl  <=  M_SCL_IN & ~m_scl_d;
        falling_m_scl <= ~M_SCL_IN &  m_scl_d;
	end
 	assign m_stop  = (~m_sda_d &  M_SDA_IN & M_SCL_IN);
    assign m_start = ( m_sda_d & ~M_SDA_IN & M_SCL_IN);
    assign s_nack  = (~s_scl_d &  S_SCL_IN & S_SDA_IN);

    /**************************************************************************/
    /* Master SCL Bit Counter                                                 */
    /*                                                                        */
    /* 9 bits = 8 bit data + ACK/NACK                                         */
    /* Clocked on falling edge of master SCL                                  */
    /**************************************************************************/
	cy_psoc3_count7 #(.cy_period(7'd8),.cy_route_ld(1),
                      .cy_route_en(1),.cy_alt_mode(0)) 
    BitCounter
    (
        /*  input          */  .clock(CLOCK),
        /*  input          */  .reset(1'b0),
        /*  input          */  .load(i2c_start),
        /*  input          */  .enable(rising_m_scl),
        /*  output [06:00] */  .count(bit_cnt),
        /*  output         */  .tc(bit_tc)
    );	

    /**************************************************************************/
    /* Master Bus Start Bit Detect                                            */
    /**************************************************************************/
	always @(posedge CLOCK) 
    begin
        /* I2C start condition = falling edge on SDA when SCL=1 */
		if (m_sda_d & ~M_SDA_IN & M_SCL_IN)
        begin
			i2c_start <= 1'b1;
        end
        /* De-assert after one master SCL clock cycle */
		else if (rising_m_scl)
		begin
			i2c_start <= 1'b0;
		end
	end

    /**************************************************************************/
    /* Master Bus Read/Write Cycle Detect                                     */
    /*                                                                        */
    /* Final bit of the 8-bit address is the read/write bit                   */
    /**************************************************************************/
    always @(posedge CLOCK) 
    begin
        if (falling_m_scl)	
        begin
            if(i2c_start)
            begin
                addr_phase <= 1'b1;
            end
            else if (bit_cnt == 7'd0)
            begin
                addr_phase <= 1'b0;
            end
        end
    end         
             
    /* Capture R/W bit during I2C address phase */
    always @(posedge CLOCK) 
    begin
        if (rising_m_scl)	
        begin
            if ((bit_cnt == 7'd2) & addr_phase)
            begin
                read_cycle <= M_SDA_IN;
            end
        end
        /* Reset if I2C Stop or repeated start detected */
        else if (m_stop | m_start)
        begin
            read_cycle <= 1'b0;
        end
    end
   
    /**************************************************************************/
    /* Determine correct time to to generate/receive ACK/NACK                 */
    /*                                                                        */
    /* Time for ACK is the last bit of the 9-bit transfer                     */
    /**************************************************************************/
    always @(posedge CLOCK) 
    begin
        if (falling_m_scl)	
        begin
            ack_time <= (bit_cnt == 7'd1);
        end
    end
    
    /**************************************************************************/
    /* Detect back-to-back data transfer                                      */
    /*                                                                        */
    /* This scenario could result in a potential clock stretch                */
    /**************************************************************************/
    always @(posedge CLOCK) 
    begin
        /* Need to detect this one clock cycle earlier than failling_m_scl */
        if (~M_SCL_IN)
        begin
            back_to_back <= (bit_cnt == 7'd0);
        end
        else
        begin
            back_to_back <= 1'b0;
        end
    end
    
    /**************************************************************************/
    /* Handle NACK to address phase on slave side                             */
    /**************************************************************************/
    always @(posedge CLOCK) 
    begin
        /* If a stop or start is detected on the master side, clear the flag */
        if (m_start | m_stop)
        begin
            addr_nack <= 1'b0;
        end
        /* NACK condition on slave side */
        else if (ack_time & addr_phase & s_nack)	
        begin
            addr_nack <= 1'b1;
        end
    end
   
    /* When slave NACK detected, wait for next falling master side SCL in */
    /* case transfer was on upstream bus */
    always @(posedge CLOCK) 
    begin
        if (addr_nack) 
        begin
            if (falling_m_scl)
            begin
                upstream_sync <= 1'b1;
            end
        end
        else
        begin
            upstream_sync <= 1'b0;
        end
    end    

    /**************************************************************************/
    /* Counter that counts the width of the master's SCL                      */
    /*                                                                        */
    /* At 100 kHz SCL, typical low time = 8 counts                            */
    /**************************************************************************/
    always @(posedge CLOCK) 
    begin
        if(falling_m_scl)
        begin
            scl_cnt <= 3'b000;
        end
        else
        begin
            /* 8 = full width of SCL low */
            if (scl_cnt != 3'b111)
            begin
                scl_cnt <= scl_cnt + 1;
            end    
        end
    end        
                
    /**************************************************************************/
    /* Logic to connect master (upstream) I2C bus to slave (downstream) I2C   */
    /* bus depending on various conditions                                    */
    /*                                                                        */
    /* Need to switch the direction of the I/O drivers depending on:          */
    /*   1) Data transfer direction (read vs. write)                          */
    /*   2) Who is generating the ACK (master or slave)                       */
    /*   3) Who is/might be stretching the clock (master or slave)            */
    /**************************************************************************/
    localparam HALF_BIT_WIDTH = 3'b011;    /* Half the width of SCL low */
    localparam FULL_BIT_WIDTH = 3'b110;    /* Full width of SCL low */
        
    /* Connect master bus to slave bus depending on various conditions */
	always @(posedge CLOCK) 
    begin
        /* If transfer is on master side or to non-existent device, hold */
        /* slave bus in IDLE state until next transfer starts */
        if (addr_nack)
        begin
            if (~upstream_sync)
            begin
                M_SCL_OUT <= 1'b1;
                M_SDA_OUT <= 1'b1;
                S_SCL_OUT <= 1'b1;
                S_SDA_OUT <= 1'b1;
            end
            else
            begin
                M_SCL_OUT <= 1'b1;
                M_SDA_OUT <= 1'b1;
                S_SCL_OUT <= M_SCL_IN;
                S_SDA_OUT <= M_SDA_IN;
            end
        end
         /* Address phase or write cycle can be handled in the same way */
        else if (addr_phase | ~read_cycle)
        begin
            /* Allow for possibility of clock stretch by slave */
            if (ack_time | (back_to_back & ~addr_phase))
            begin
                /* Force a stretch on both clocks for 1/2 SCL period */
                if (scl_cnt <= HALF_BIT_WIDTH)
                begin
                    M_SCL_OUT <= 1'b0;
                    M_SDA_OUT <= 1'b1;
                    S_SCL_OUT <= 1'b0;
                    /* Propagate master ACK down to slave when appropriate */
                    if (ack_time)
                    begin
                        S_SDA_OUT <= 1'b1;
                    end
                    else
                    begin
                        S_SDA_OUT <= M_SDA_IN;
                    end
                end
                else if (scl_cnt <= FULL_BIT_WIDTH)
                /* Connect slave ACK/NACK to master */
                begin
                    if (ack_time)
                    begin
                        M_SCL_OUT <= S_SCL_IN;
                        M_SDA_OUT <= S_SDA_IN;
                        S_SDA_OUT <= 1'b1;
                        S_SCL_OUT <= M_SCL_IN;
                    end
                    else
                    /* (back_to_back & ~addr_phase) */
                    begin
                        M_SCL_OUT <= S_SCL_IN;
                        M_SDA_OUT <= 1'b1;
                        S_SCL_OUT <= M_SCL_IN;
                        S_SDA_OUT <= M_SDA_IN;
                    end
                end
                else
                begin
                    /* Propagate slave ACK up to master when appropriate */
                    if (ack_time)
                    begin
                        M_SCL_OUT <= S_SCL_IN;
                        M_SDA_OUT <= S_SDA_IN;
                        S_SCL_OUT <= 1'b1;
                        S_SDA_OUT <= 1'b1;
                    end
                    else
                    /* (back_to_back & ~addr_phase) */
                    begin
                        M_SCL_OUT <= 1'b1;
                        M_SDA_OUT <= 1'b1;
                        S_SCL_OUT <= M_SCL_IN;
                        S_SDA_OUT <= M_SDA_IN;
                    end
                end
            end
            else
            /* Not ACK or back-to-back time, so connect master to slave */
            begin
                M_SCL_OUT <= 1'b1;
                M_SDA_OUT <= 1'b1;
                S_SCL_OUT <= M_SCL_IN;
                S_SDA_OUT <= M_SDA_IN;
            end
        end
         /* Handle Master Read Cycles */
        else if (read_cycle)
        begin
            if (ack_time | (back_to_back & ~addr_phase))
            begin
                /* Force a stretch on both clocks for 1/2 SCL period */
                if (scl_cnt <= HALF_BIT_WIDTH)
                begin
                    M_SCL_OUT <= 1'b0;
                    M_SDA_OUT <= 1'b1;
                    S_SCL_OUT <= 1'b0;
                    S_SDA_OUT <= M_SDA_IN;
                end
                else if (scl_cnt <= FULL_BIT_WIDTH)
                /* Propagate master ACK down to slave when appropriate */
                begin
                    if (ack_time)
                    begin
                        M_SCL_OUT <= 1'b1;
                        M_SDA_OUT <= 1'b1;
                        S_SCL_OUT <= M_SCL_IN;
                        S_SDA_OUT <= M_SDA_IN;
                    end
                    else
                    /* (back_to_back & ~addr_phase) */
                    begin
                        M_SCL_OUT <= S_SCL_IN;
                        M_SDA_OUT <= 1'b1;
                        S_SCL_OUT <= M_SCL_IN;
                        S_SDA_OUT <= M_SDA_IN;
                    end
                end
                else
                begin
                    /* Propagate master ACK down to slave when appropriate */
                    if (ack_time)
                    begin
                        M_SCL_OUT <= 1'b1;
                        M_SDA_OUT <= 1'b1;
                        S_SCL_OUT <= M_SCL_IN;
                        S_SDA_OUT <= M_SDA_IN;
                    end
                    else
                    /* (back_to_back & ~addr_phase) */
                    begin
                        M_SCL_OUT <= S_SCL_IN;
                        M_SDA_OUT <= S_SDA_IN;  
                        S_SCL_OUT <= 1'b1;
                        S_SDA_OUT <= M_SDA_IN;
                    end
                end
            end    
            else
            /* Not ACK or back-to-back time, so connect slave to master */
            begin
                M_SCL_OUT <= 1'b1;
                M_SDA_OUT <= S_SDA_IN;  
                S_SCL_OUT <= M_SCL_IN;
                S_SDA_OUT <= 1'b1;
            end
        end        
    end
endmodule
